#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

__global__ void optimized_module_fn_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,           // batch_size
    int inF,         // number of input features
    int outF,        // number of output features
    int kernel_size,
    float scale_factor
) {
    int b = blockIdx.x;
    if (b >= B) return;

    // Shared memory layout
    extern __shared__ float smem[];
    float* shared_x = smem;                        // size: inF
    float* local_out = shared_x + inF;            // size: outF
    float* partial_sums = local_out + outF;       // size: blockDim.x

    // Cooperative loading of input vector into shared memory
    for (int i = threadIdx.x; i < inF; i += blockDim.x) {
        shared_x[i] = x[b * inF + i];
    }
    __syncthreads();

    // Compute linear transformation using warp-level parallelism
    #pragma unroll 4
    for (int j = threadIdx.x; j < outF; j += blockDim.x) {
        float val = bias[j];
        for (int i = 0; i < inF; i++) {
            val = __fmaf_rn(shared_x[i], weight[j * inF + i], val);
        }
        local_out[j] = val;
    }
    __syncthreads();

    // Max pooling with improved memory access pattern
    int pooled_len = (outF >= kernel_size) ? (1 + (outF - kernel_size) / kernel_size) : 0;
    
    float sum_local = 0.0f;
    #pragma unroll 2
    for (int seg = threadIdx.x; seg < pooled_len; seg += blockDim.x) {
        int start = seg * kernel_size;
        float m_val = local_out[start];
        
        #pragma unroll
        for (int offset = 1; offset < kernel_size; offset++) {
            m_val = fmaxf(m_val, local_out[start + offset]);
        }
        sum_local += m_val;
    }
    partial_sums[threadIdx.x] = sum_local;
    __syncthreads();

    // Warp-optimized reduction
    if (threadIdx.x < 32) {
        #pragma unroll
        for (int offset = blockDim.x/2; offset >= 32; offset >>= 1) {
            partial_sums[threadIdx.x] += partial_sums[threadIdx.x + offset];
            __syncthreads();
        }
        
        // Warp-level reduction (no sync needed within a warp)
        if (threadIdx.x < 16) partial_sums[threadIdx.x] += partial_sums[threadIdx.x + 16];
        if (threadIdx.x < 8)  partial_sums[threadIdx.x] += partial_sums[threadIdx.x + 8];
        if (threadIdx.x < 4)  partial_sums[threadIdx.x] += partial_sums[threadIdx.x + 4];
        if (threadIdx.x < 2)  partial_sums[threadIdx.x] += partial_sums[threadIdx.x + 2];
        if (threadIdx.x < 1)  partial_sums[threadIdx.x] += partial_sums[threadIdx.x + 1];
    }

    if (threadIdx.x == 0) {
        output[b] = partial_sums[0] * scale_factor;
    }
}

at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    const auto B = x.size(0);
    const auto inF = x.size(1);
    const auto outF = weight.size(0);
    
    auto out = torch::empty({B}, x.options());

    const int threads = 256;
    size_t sharedMemSize = (inF + outF + threads) * sizeof(float);

    optimized_module_fn_kernel<<<B, threads, sharedMemSize>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        (int)kernel_size,
        (float)scale_factor
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Optimized CUDA forward implementation");
}